// -*- C++ -*-

/**
 * @file alignedmmap.h
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */


#ifndef _ALIGNEDMMAP_H_
#define _ALIGNEDMMAP_H_

#if defined(_WIN32)
//#error "You should not include this header on Windows."
#endif

#include <map>

#include "sassert.h"
#include "stlallocator.h"
#include "freelistheap.h"

using namespace std;
using namespace HL;


#include "mmapwrapper.h"

/**
 * @class MmapAlloc
 * @brief Obtains memory from Mmap but doesn't allow it to be freed.
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */

class MmapAlloc {
public:
  void * malloc (size_t sz) {
    void * ptr = MmapWrapper::map (sz);
    return ptr;
  }

};


/**
 * @class BumpAlloc
 * @brief Obtains memory in chunks and bumps a pointer through the chunks.
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */

template <int ChunkSize,
	  class super>
class BumpAlloc : public super {
public:

  BumpAlloc (void)
    : _bump (NULL),
      _remaining (0)
  {}

  inline void * malloc (size_t sz) {
    // If there's not enough space left to fulfill this request, get
    // another chunk.
    if (_remaining < sz) {
      refill(sz);
    }
    char * old = _bump;
    _bump += sz;
    _remaining -= sz;
    return old;
  }

  /// Free is disabled (we only bump, never reclaim).
  inline void free (void *) {}

private:

  /// The bump pointer.
  char * _bump;

  /// How much space remains in the current chunk.
  size_t _remaining;

  // Get another chunk.
  void refill (size_t sz) {
    if (sz < ChunkSize) {
      sz = ChunkSize;
    }
    _bump = (char *) super::malloc (sz);
    _remaining = sz;
  }

};


/**
 * @class ExactlyOne
 * @brief Creates a singleton of type CLASS, accessed through ().
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */


template <class CLASS>
class ExactlyOne {
public:
  
  inline CLASS& operator() (void) {
    static CLASS theOneTrueInstance;
    return theOneTrueInstance;
  }

};

/**
 * @class AlignedMmapInstance
 * @brief Memory allocated from here is aligned with respect to Alignment.
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */


template <size_t Alignment_>
class AlignedMmapInstance {
public:

  enum { Alignment = Alignment_ };

  inline void * malloc (size_t sz) {

    // Get memory from mmap,
    // align a pointer in the space,
    // and free the space before and after the aligned segment.

    char * ptr = reinterpret_cast<char *>(MmapWrapper::map (sz + Alignment));
    char * newptr = align (ptr);

    //
    // Unmap the part before (prolog) and after (prolog).
    //
    // prolog = extra space before the aligned part (PR below).
    // Below is a diagram of the allocated space.
    // p = ptr, n = newptr, EP = epilog.

    // p  n
    // ^  ^-----^  ^
    //  PR  sz   EP

    size_t prolog = (size_t) newptr - (size_t) ptr;

    if (prolog >= Alignment) {
      MmapWrapper::unmap (ptr, prolog);
    }

    // Now unmap the part after.
    // eob = end of buffer (the pointer just after EP).

    void * eob = ptr + Alignment + sz - 1;
    void * originalStart = newptr + sz;
    void * newStart = (void *) align ((char *) originalStart);

    if ((size_t) newStart <= (size_t) eob) {
      size_t epilog = (size_t) eob - (size_t) newStart;
      if (epilog >= Alignment) {
	MmapWrapper::unmap (newStart, epilog);
      }
    }

    // Now record the size (rounded up to the next page) associated
    // with this pointer.

    MyMap[newptr] = (valType) align ((char *) sz);

    assert (MyMap.find(newptr) != MyMap.end());
    assert (reinterpret_cast<size_t>(newptr) % Alignment == 0);
    return newptr;
  }

  inline void free (void * ptr) {

    // Find the object. If we don't find it, we didn't allocate it.
    // For now, just ignore such an invalid free...

    typename mapType::iterator i = MyMap.find (ptr);
    if (i == MyMap.end()) {
      return;
    }

    // Now unmap the original range corresponding to this pointer.
    size_t requestedSize = (*i).second;
    MmapWrapper::unmap (ptr, requestedSize);

    // Finally, undo the mapping.
    MyMap.erase (i);
  }
  
  inline size_t getSize (void * ptr) {
    typename mapType::iterator i;
    i = MyMap.find (ptr);
    if (i == MyMap.end()) {
      // If we didn't allocate this object, pretend that it was of size 0.
      return 0;
    } else {
      size_t requestedSize = (*i).second;
      return requestedSize;
    }
  }


private:

  /// Round a pointer up to the next Alignment-rounded chunk.
  inline static char * align (char * buf) {
    return (char *)(((size_t) buf + (Alignment-1)) & ~(Alignment-1));
  }

  // Manage information in an STL map that uses a custom heap for
  // allocation.

  /// The key is an mmapped pointer.
  typedef void * keyType;

  /// The value is the requested size.
  typedef size_t valType;

  // The heap from which memory comes for the Map's purposes.
  // Objects come from chunks via mmap, and we manage these with a free list.
  class SourceHeap : public FreelistHeap<BumpAlloc<16384, MmapAlloc> > {};

  /// A custom STL-compatible adapter for the SourceHeap.
  template <class OBJTYPE>
  class CustomAllocator : public STLAllocator<OBJTYPE, SourceHeap> {};

  /// The map type, with all the pieces in place.
  typedef std::map<keyType, valType,
	      std::less<keyType>,
	      CustomAllocator<pair<keyType, valType> > > mapType;

  /// The map that maintains the size of each mmapped chunk.
  mapType MyMap;

};


/**
 * @class AlignedMmap
 * @brief Route requests to the one aligned mmap instance.
 * @author Emery Berger <http://www.cs.umass.edu/~emery>
 */

template <size_t Alignment_>
class AlignedMmap : public ExactlyOne<AlignedMmapInstance<Alignment_> > {
public:

  enum { Alignment = Alignment_ };

  inline void * malloc (size_t sz) {
    return (*this)().malloc (sz);
  }
  inline void free (void * ptr) {
    return (*this)().free (ptr);
  }
  inline size_t getSize (void * ptr) const {
    return (*this)().getSize (ptr);
  }
 
};


#endif
